#!/usr/bin/env python3
#
# Copyright Contributors to the fusa-gcc-plugin project
#
# SPDX-License-Identifier: LGPL-2.1-or-later

import argparse
import json
import logging
import random
import shutil
import subprocess
import sys
import tempfile
import time

from dataclasses import dataclass
from pathlib import Path
from typing import Collection, Iterable, Optional


class APISetLoadError(Exception):
    """Error state for loading an api-set"""


def generate_apis_json(
    plugin_path: Path, api_sets_dir: Path, profiles: list[str], output_dir: Path
) -> Path:
    """Generate the valid_apis.json file based on api-sets input"""
    api_jsons = process_api_sets(plugin_path, api_sets_dir, output_dir)

    final_json_file = output_dir / "valid_apis.json"
    merge_api_jsons(api_jsons, profiles, final_json_file)

    return final_json_file


def process_api_sets(
    plugin_path: Path, api_sets_dir: Path, output_dir: Path
) -> list[Path]:
    """Process api-sets directory to generate valid_apis.json"""
    print(f"Loading api-sets from: {api_sets_dir}")
    json_files = []
    for subdir in api_sets_dir.iterdir():
        if not subdir.is_dir():
            continue
        print(f"Loading api-set: {subdir}")
        try:
            list_file, template_file, cflags_file = load_api_set(subdir)
        except APISetLoadError:
            continue

        print(f"Processing api-set: {subdir}")
        json_files.extend(
            process_api_set(
                plugin_path, list_file, template_file, cflags_file, output_dir
            )
        )

    return json_files


def load_api_set(subdir: Path) -> tuple[Path, Path, Optional[Path]]:
    """Load api-set subdirectory safely"""
    list_file = subdir / "list"
    template_file = None
    cflags_file = subdir / "cflags"

    try:
        template_file = next(subdir.glob("template.*"))
    except StopIteration:
        print(f"Warning: No template file found for api-set: {subdir}")
        raise APISetLoadError()

    if not list_file.exists():
        print(f"Warning: No list file found for api-set: {subdir}")
        raise APISetLoadError()

    if not cflags_file.exists():
        print(f"Note: No cflags file found for api-set: {subdir}")
        cflags_file = None

    return list_file, template_file, cflags_file


def process_api_set(
    plugin_path: Path,
    list_file: Path,
    template_file: Path,
    cflags_file: Optional[Path],
    output_dir: Path,
) -> list[Path]:
    """Process subdirectory of api-sets to generate C files"""
    print(f"Loading cflags from {cflags_file}")
    cflags = []
    if cflags_file:
        cflags_str = cflags_file.read_text()
        cflags = cflags_str.split()

    print(f"Processing {list_file} using {template_file}")
    json_files = []
    api_list = list_file.read_text(encoding="utf-8")
    for api in api_list.splitlines():
        api_scope, api_func = api.split(",", 1)
        synthetic_file = generate_api_synthetic_file(
            api_scope,
            api_func,
            template_file,
            output_dir / "synthetic-files" / api_scope,
        )
        json_file = process_synthetic_file(
            plugin_path, synthetic_file, cflags, output_dir / "json-files" / api_scope
        )
        if json_file:
            json_files.append(json_file)
        else:
            print(f"Error: could not generate json for {api_scope}/{api_func}")

    return json_files


def generate_api_synthetic_file(
    api_scope, api_func, template_file: Path, output_dir
) -> Path:
    """Generate synthetic file for API to feed into fusa-gcc-plugin"""
    api_template = template_file.read_text(encoding="utf-8")
    output_dir.mkdir(parents=True, exist_ok=True)
    output_file = output_dir / f"test-{api_scope}-{api_func}{template_file.suffix}"
    print(f"Processing {api_scope}/{api_func} -> {output_file}")
    api_c_file = api_template.replace("FUNCTION", api_func)
    output_file.write_text(api_c_file)
    return output_file


def process_synthetic_file(
    plugin_path: Path, c_file: Path, cflags: Iterable[str], output_dir: Path
) -> Optional[Path]:
    """Run fusa-gcc-plugin on a synthetic file to dump associated apis"""
    print(f"Processing c-file: {c_file}")
    c_file_fullpath = c_file.resolve()
    output_dir.mkdir(parents=True, exist_ok=True)
    work_dir = output_dir / "tmp"
    work_dir.mkdir(parents=True, exist_ok=True)
    plugin_dump_apis_cmd = [
        "gcc",
        f"-fplugin={plugin_path}",
    ]
    plugin_dump_apis_cmd.extend(cflags)
    plugin_dump_apis_cmd.extend(
        [
            "-c",
            str(c_file_fullpath),
            f"-fplugin-arg-fusa-dump-apis=.",
        ]
    )
    print(" ".join(plugin_dump_apis_cmd))
    dump_proc = subprocess.run(
        plugin_dump_apis_cmd,
        cwd=work_dir,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        check=False,
    )
    print(dump_proc.returncode)
    output = dump_proc.stdout.decode("utf-8")
    print(output)
    if dump_proc.returncode != 0:
        print(f"Error: could not process: {c_file}")
        sys.exit(1)
    if "Symbol database created at: " in output:
        _, json_file = output.rsplit(" ", 1)
        json_file = work_dir / json_file.strip()
        json_file.rename(output_dir / json_file.name)
        return output_dir / json_file.name
    else:
        return None


def merge_api_jsons(
    api_json_files: list[Path], profiles: list[str], output_json_file: Path
) -> None:
    """Merge json files per api into a single json"""
    print(f"Merging api json files...")
    output_json = {"version": 1, "certified_symbols": []}
    for api_json_file in api_json_files:
        print(f"Merging {api_json_file}")
        with open(api_json_file) as api_json_f:
            api_json = json.load(api_json_f)
        # Strip top level json structure to get to the juicy bits
        api_json = api_json["symbols"][0]
        # Transformations that should probably be in the actual plugin
        api_json["symbol"] = api_json["name"]
        del api_json["name"]
        del api_json["class_hierarchy"]
        del api_json["namespace_hierarchy"]
        # Inject the profiles here
        api_json["profiles"] = profiles
        # Append
        output_json["certified_symbols"].append(api_json)

    with open(output_json_file, "w") as output_json_f:
        json.dump(output_json, output_json_f, indent=4)


def fusa_plugin_available(so_path: Path) -> bool:
    """Execute gcc to check if the fusa plugin is available"""
    tmp_file = tempfile.NamedTemporaryFile(suffix=".c")
    test_c_file = Path(tmp_file.name)
    test_c_file.write_text("int main() {}")
    gcc_cmd = ["gcc", f"-fplugin={so_path}", f"{test_c_file}", "-fplugin-arg-fusa-usage"]
    print(f"Note: Executing: {' '.join(gcc_cmd)}")
    test_gcc = subprocess.run(
        gcc_cmd,
        cwd=str(test_c_file.parent),
        check=False,
    )
    test_c_file.unlink()
    return test_gcc.returncode == 0


def prepare_output_dir(output_dir: Path) -> bool:
    if output_dir.exists():
        print(f"Output directory already exists: {output_dir}")
        max_description_items = 10
        directory_description = [
            f"  {p}"
            for _, p in zip(range(max_description_items), output_dir.rglob("*"))
        ]
        directory_description.append("  ...")
        print("Preview of directory:")
        print("\n".join(directory_description))
        confirmation = input("Delete existing files [Y/n]? ")
        if confirmation.lower().startswith("y"):
            shutil.rmtree(output_dir)
        else:
            return False

    output_dir.mkdir(parents=True)
    return True


def build_argparser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="fusa-gcc-plugin-generator",
        description="Synthesizer for valid_apis.json")
    parser.add_argument(
        "--plugin-path",
        help="File path for fusa plugin so file (/usr/lib64/fusa.so)",
        type=Path,
        default=Path("/usr/lib64/fusa.so"),
    )
    parser.add_argument(
        "--api-sets",
        help="Root directory of APIs to synthesize (./api-sets)",
        type=Path,
        default=Path("./api-sets"),
    )
    parser.add_argument(
        "--profile",
        help="ASIL Profile of output json",
        choices=["asil-b"],
        default="asil-b",
    )
    parser.add_argument(
        "--output",
        help="Output directory of results (./output)",
        type=Path,
        default=Path("./output"),
    )
    return parser


def generate():
    # Parse arguments to script
    parser = build_argparser()
    args = parser.parse_args()
    plugin_path = args.plugin_path
    api_sets_dir = args.api_sets.resolve()
    output_profile = args.profile
    output_dir = args.output / f"{time.time()}"

    # Print pretty banner
    print("Starting synthesizer")

    # Check api_sets_dir is directory
    if not api_sets_dir.is_dir():
        print(f"Error: api-sets directory isn't a directory ({api_sets_dir})")
        sys.exit(1)

    # Check fusa-gcc-plugin is installed
    print("Checking for gcc fusa plugin")
    if not fusa_plugin_available(plugin_path):
        print("Error: Fusa plugin not found")
        sys.exit(1)
    print("Done!")

    # Clean output directory if needed
    if not prepare_output_dir(output_dir):
        print("Script aborting.")
        sys.exit(1)

    # Synthesize the C files that call the api-sets functions
    final_json_file = generate_apis_json(
        plugin_path, api_sets_dir, output_profile, output_dir
    )

    def sizeof_fmt(num, suffix="B"):
        for unit in ("", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"):
            if abs(num) < 1024.0:
                return f"{num:3.1f}{unit}{suffix}"
            num /= 1024.0
        return f"{num:.1f}Yi{suffix}"

    final_json_file_size = sizeof_fmt(final_json_file.stat().st_size)
    print(f"\nOutput: {final_json_file} {final_json_file_size}")


if __name__ == "__main__":
    generate()
